"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.BullmqBackend = void 0;
const lodash_1 = __importDefault(require("lodash"));
const bullmq_1 = require("bullmq");
const debug_1 = __importDefault(require("debug"));
const worker_1 = require("./worker");
const debug = (0, debug_1.default)('hyperworker:backend');
const sleep = (t) => new Promise((resolve, _reject) => setTimeout(() => resolve(true), t));
const defaultWorkerOpts = {
    lockDuration: 30000,
};
const defaultQueueOpts = {
    // make data sizes predictable, modest, and auto healing
    streams: {
        events: { maxLen: 1000 }, // default was 10k
    },
    defaultJobOptions: {
        removeOnComplete: true,
        removeOnFail: { count: 100 },
        stackTraceLimit: 5, // default no limit
    },
};
class BullmqBackend {
    constructor(workers, connection, opts) {
        this.connection = connection;
        this.workers = workers;
        this.opts = {
            workerOpts: lodash_1.default.merge({}, defaultWorkerOpts, opts === null || opts === void 0 ? void 0 : opts.workerOpts),
            queueOpts: lodash_1.default.merge({}, defaultQueueOpts, opts === null || opts === void 0 ? void 0 : opts.queueOpts),
        };
        debug('opts: %o', this.opts);
        this.bmqueues = workers.map((Klass) => {
            debug(`adding queue: ${Klass.queueName} from class ${Klass.name}`);
            const q = new bullmq_1.Queue(Klass.queueName, Object.assign({ connection: this.connection }, this.opts.queueOpts));
            q.on('error', (err) => {
                worker_1.HyperWorker.logger && worker_1.HyperWorker.logger.error(`[error] ${err}`);
                worker_1.HyperWorker.reportError(err);
            });
            const enqueue = (data, opts) => q.add('job', data, opts);
            Klass.enqueue = enqueue;
            Klass.queue = q;
            return q;
        });
        this.workers = workers;
    }
    processJobs() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.bmworkers) {
                throw new Error('already processing jobs');
            }
            if (!worker_1.HyperWorker.logger) {
                throw new Error('logger is missing. did the app boot correctly?');
            }
            const { magenta, bold } = worker_1.HyperWorker.logger.colors;
            const pref = bold(magenta('workers:'));
            if (!this.workers) {
                throw new Error("you need to wire into your worker first. call 'wireInto(<workers>)'");
            }
            this.bmworkers = this.workers
                .filter((Klass) => {
                if (Klass.producerOnly) {
                    worker_1.HyperWorker.logger.info(`${pref} producer only job: ${Klass.queueName}`);
                    return false;
                }
                return true;
            })
                .map((Klass) => {
                debug(`processing queue: ${Klass.queueName} from class ${Klass.name}`);
                return new bullmq_1.Worker(Klass.queueName, (job) => __awaiter(this, void 0, void 0, function* () {
                    debug(`processing job in: ${Klass.queueName}`);
                    const w = new Klass();
                    yield w.perform(job);
                }), Object.assign({ connection: this.connection }, this.opts.workerOpts));
            });
            this.bmworkers.forEach((w) => {
                w.on('drained', (_job) => {
                    worker_1.HyperWorker.logger.info(`${pref} drained: ${w.name}`);
                });
                w.on('completed', (_job) => {
                    worker_1.HyperWorker.logger.info(`${pref} completed: ${w.name}`);
                });
                w.on('failed', (job, err) => {
                    worker_1.HyperWorker.logger.info(`${pref} failed: ${w.name} (${err})`);
                    worker_1.HyperWorker.reportError(err, { worker: w.name, jobId: job.id });
                });
                w.on('error', (err) => {
                    worker_1.HyperWorker.logger.error(`${pref} error: ${w.name} (${err})`);
                    worker_1.HyperWorker.reportError(err, { worker: w.name });
                });
            });
        });
    }
    stop() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.bmworkers) {
                for (const w of this.bmworkers) {
                    if (w) {
                        w.connection.removeAllListeners('close');
                        w.connection.removeAllListeners('error');
                        w.removeAllListeners('error');
                        w.removeAllListeners('completed');
                        w.removeAllListeners('drained');
                        yield w.close();
                    }
                }
            }
            if (this.bmqueues) {
                for (const q of this.bmqueues) {
                    if (q) {
                        q.connection.removeAllListeners('close');
                        q.connection.removeAllListeners('error');
                        q.removeAllListeners('error');
                        q.removeAllListeners('completed');
                        q.removeAllListeners('drained');
                        yield q.close();
                    }
                }
            }
            yield this.connection.quit();
            this.workers = undefined;
            yield sleep(500); // give some time for emitters to drain
        });
    }
}
exports.BullmqBackend = BullmqBackend;
