"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.buildWorkers = exports.createConnection = exports.loadWorkers = exports.createBackend = void 0;
/* eslint-disable @typescript-eslint/no-var-requires */
const glob_1 = require("glob");
const lodash_1 = __importDefault(require("lodash"));
const debug_1 = __importDefault(require("debug"));
const worker_1 = require("./worker");
const mailer_1 = require("./mailer");
const backend_1 = require("./backend");
const debug = (0, debug_1.default)('@hyperstackjs/hyperworker');
const load = (patt) => (0, glob_1.globSync)(patt).map((f) => {
    let mod = null;
    try {
        mod = require(f);
    }
    catch (e) {
        console.log(`error dynamically loading '${f}'\n${e.toString()}`); // eslint-disable-line no-console
        throw e;
    }
    return mod.default || mod;
});
const createBackend = ({ hyperworkers, redis, backendOpts, }) => {
    return new backend_1.BullmqBackend(hyperworkers, redis, backendOpts);
};
exports.createBackend = createBackend;
const loadWorkers = ({ workersGlob, mailersGlob }) => {
    const workers = lodash_1.default.reduce(load(workersGlob), lodash_1.default.merge) || {};
    const mailers = lodash_1.default.reduce(load(mailersGlob), lodash_1.default.merge) || {};
    if (Object.keys(workers).includes('default') ||
        Object.keys(mailers).includes('default')) {
        throw new Error('cannot start worker that is exported as default');
    }
    return { workers, mailers };
};
exports.loadWorkers = loadWorkers;
const createConnection = ({ redisUrl, rejectUnauthorized = false, }) => __awaiter(void 0, void 0, void 0, function* () {
    const Redis = require('ioredis');
    return new Redis(redisUrl, {
        name: 'hyperworkers',
        lazyConnect: true,
        maxRetriesPerRequest: null,
        enableReadyCheck: false,
        tls: (redisUrl === null || redisUrl === void 0 ? void 0 : redisUrl.startsWith('rediss://'))
            ? {
                rejectUnauthorized,
            }
            : undefined,
    });
});
exports.createConnection = createConnection;
const initializeWorkers = ({ logger, mailerTemplatesRoot, mailerSettings, }) => {
    worker_1.HyperWorker.clear();
    mailer_1.Mailer.clear();
    worker_1.HyperWorker.initWorker({ logger });
    mailer_1.Mailer.initMailer({
        templatesRoot: mailerTemplatesRoot,
        settings: mailerSettings,
    });
};
// recv: beforeWorkers, afterWorkers (like controllers)
const buildWorkers = ({ isInProcess, workersGlob, mailersGlob, mailerTemplatesRoot, mailerSettings, logger, redisUrl, backendOpts, truncate = false, beforeWorkers = (_p) => { }, afterWorkers = (_p) => { }, }, mode) => __awaiter(void 0, void 0, void 0, function* () {
    // worker types rigging
    initializeWorkers({ logger, mailerTemplatesRoot, mailerSettings });
    // set up mail transport and options
    // set up hyperworker logger, report errors with sentry
    beforeWorkers({ HyperWorker: worker_1.HyperWorker, Mailer: mailer_1.Mailer });
    // load workers from disk
    const { workers, mailers } = (0, exports.loadWorkers)({
        workersGlob,
        mailersGlob,
    });
    const numWorkers = Object.keys(workers).length;
    const numMailers = Object.keys(mailers).length;
    const workerline = `${numWorkers} workers, ${numMailers} mailers`;
    const { magenta, bold } = logger.colors;
    const pref = bold(magenta('workers:'));
    if (isInProcess) {
        debug('running in-process');
        logger.info(`${pref} [in-process, duplex] ${workerline}`);
        worker_1.HyperWorker.setMode({ inprocess: isInProcess });
        return {
            background: 'inprocess',
            backend: null,
            connection: null,
            mode: 'duplex',
            mailerTemplatesRoot,
            numMailers,
            numWorkers,
            workers,
            mailers,
        };
    }
    const hyperworkers = [
        ...Object.values(workers),
        ...Object.values(mailers),
    ];
    // set up a redis connection
    if (!redisUrl) {
        throw new Error('cannot find a Redis URL');
    }
    debug('creating a Redis connection');
    const redis = yield (0, exports.createConnection)({ redisUrl });
    yield redis.connect();
    debug('connected to Redis');
    logger.info(`${pref} [queue, ${mode}] ${workerline}`);
    if (truncate) {
        debug('truncating Redis content');
        logger.warn(`${pref} truncating Redis contents`);
        yield redis.flushdb();
    }
    // build the backend and finish
    const b = (0, exports.createBackend)({ hyperworkers, redis, backendOpts });
    debug('created backend');
    // share the backend for message passing, control, or in Redis
    // case, temporary store for workers: backend.connection
    // this is a bit of an oversharing, but very pragmatic
    worker_1.HyperWorker.backend = b;
    afterWorkers({ backend: b });
    if (mode === 'consumer' || mode === 'duplex') {
        debug('processing jobs: consumer mode');
        b.processJobs();
    }
    // afterWorkers({ backend })
    return {
        background: 'queue',
        backend: b,
        connection: redis,
        mailerTemplatesRoot,
        mode,
        numMailers,
        numWorkers,
        workers,
        mailers,
    };
});
exports.buildWorkers = buildWorkers;
